@@ -1,19 +1,21 @@
-// WhoAmI
-// Copyright © 2017-2021 Jeron Aldaron Lau.
+// Copyright © 2017-2022 The WhoAmI Contributors.
 //
 // Licensed under any of:
-//  - Apache License, Version 2.0 (https://www.apache.org/licenses/LICENSE-2.0)
-//  - MIT License (https://mit-license.org/)
-//  - Boost Software License, Version 1.0 (https://www.boost.org/LICENSE_1_0.txt)
+// - Apache License, Version 2.0 (https://www.apache.org/licenses/LICENSE-2.0)
+// - Boost Software License, Version 1.0 (https://www.boost.org/LICENSE_1_0.txt)
+// - MIT License (https://mit-license.org/)
 // At your choosing (See accompanying files LICENSE_APACHE_2_0.txt,
 // LICENSE_MIT.txt and LICENSE_BOOST_1_0.txt).
 
-use crate::{DesktopEnv, Platform};
-
-use std::ffi::{c_void, OsString};
-use std::mem;
-use std::os::unix::ffi::OsStringExt;
-
+use std::{
+    borrow::Cow,
+    ffi::{c_void, CStr, OsString},
+    mem,
+    os::{
+        raw::{c_char, c_int},
+        unix::ffi::OsStringExt,
+    },
+};
 #[cfg(target_os = "macos")]
 use std::{
     os::{
@@ -23,6 +25,8 @@ use std::{
     ptr::null_mut,
 };
 
+use crate::{Arch, DesktopEnv, Platform};
+
 #[repr(C)]
 struct PassWd {
     pw_name: *const c_void,
@@ -174,13 +178,13 @@ fn os_from_cfstring(string: *mut c_void) -> OsString {
     unsafe {
         let len = CFStringGetLength(string);
         let capacity =
-            CFStringGetMaximumSizeForEncoding(len, 134_217_984 /*UTF8*/) + 1;
+            CFStringGetMaximumSizeForEncoding(len, 134_217_984 /* UTF8 */) + 1;
         let mut out = Vec::with_capacity(capacity as usize);
         if CFStringGetCString(
             string,
             out.as_mut_ptr(),
             capacity,
-            134_217_984, /*UTF8*/
+            134_217_984, /* UTF8 */
         ) != 0
         {
             out.set_len(strlen(out.as_ptr().cast())); // Remove trailing NUL byte
@@ -240,11 +244,11 @@ fn getpwuid(real: bool) -> OsString {
     }
 }
 
-pub fn username() -> String {
+pub(crate) fn username() -> String {
     string_from_os(username_os())
 }
 
-pub fn username_os() -> OsString {
+pub(crate) fn username_os() -> OsString {
     getpwuid(false)
 }
 
@@ -290,21 +294,21 @@ fn fancy_fallback_os(result: Result<OsString, OsString>) -> OsString {
     }
 }
 
-pub fn realname() -> String {
+pub(crate) fn realname() -> String {
     string_from_os(realname_os())
 }
 
-pub fn realname_os() -> OsString {
+pub(crate) fn realname_os() -> OsString {
     getpwuid(true)
 }
 
 #[cfg(not(target_os = "macos"))]
-pub fn devicename_os() -> OsString {
+pub(crate) fn devicename_os() -> OsString {
     devicename().into()
 }
 
-#[cfg(not(target_os = "macos"))]
-pub fn devicename() -> String {
+#[cfg(not(any(target_os = "macos", target_os = "illumos")))]
+pub(crate) fn devicename() -> String {
     let mut distro = String::new();
 
     if let Ok(program) = std::fs::read_to_string("/etc/machine-info") {
@@ -326,12 +330,12 @@ pub fn devicename() -> String {
 }
 
 #[cfg(target_os = "macos")]
-pub fn devicename() -> String {
+pub(crate) fn devicename() -> String {
     string_from_os(devicename_os())
 }
 
 #[cfg(target_os = "macos")]
-pub fn devicename_os() -> OsString {
+pub(crate) fn devicename_os() -> OsString {
     let out = os_from_cfstring(unsafe {
         SCDynamicStoreCopyComputerName(null_mut(), null_mut())
     });
@@ -344,11 +348,23 @@ pub fn devicename_os() -> OsString {
     fancy_fallback_os(computer)
 }
 
-pub fn hostname() -> String {
+#[cfg(target_os = "illumos")]
+pub(crate) fn devicename() -> String {
+    let mut nodename = String::new();
+
+    if let Ok(program) = std::fs::read_to_string("/etc/nodename") {
+        let program = program.into_bytes();
+        nodename.push_str(&String::from_utf8_lossy(&program));
+        nodename.pop(); // Remove the trailing newline
+    }
+    fancy_fallback(Err(hostname()))
+}
+
+pub(crate) fn hostname() -> String {
     string_from_os(hostname_os())
 }
 
-pub fn hostname_os() -> OsString {
+fn hostname_os() -> OsString {
     // Maximum hostname length = 255, plus a NULL byte.
     let mut string = Vec::<u8>::with_capacity(256);
     unsafe {
@@ -405,20 +421,18 @@ fn distro_xml(data: String) -> Option<String> {
         } else {
             Some(product_name.to_string())
         }
-    } else if let Some(user_visible_version) = user_visible_version {
-        Some(format!("Mac OS (Unknown) {}", user_visible_version))
     } else {
-        None
+        user_visible_version.map(|v| format!("Mac OS (Unknown) {}", v))
     }
 }
 
 #[cfg(target_os = "macos")]
-pub fn distro_os() -> Option<OsString> {
+pub(crate) fn distro_os() -> Option<OsString> {
     distro().map(|a| a.into())
 }
 
 #[cfg(target_os = "macos")]
-pub fn distro() -> Option<String> {
+pub(crate) fn distro() -> Option<String> {
     if let Ok(data) = std::fs::read_to_string(
         "/System/Library/CoreServices/ServerVersion.plist",
     ) {
@@ -433,12 +447,12 @@ pub fn distro() -> Option<String> {
 }
 
 #[cfg(not(target_os = "macos"))]
-pub fn distro_os() -> Option<OsString> {
+pub(crate) fn distro_os() -> Option<OsString> {
     distro().map(|a| a.into())
 }
 
 #[cfg(not(target_os = "macos"))]
-pub fn distro() -> Option<String> {
+pub(crate) fn distro() -> Option<String> {
     let mut distro = String::new();
 
     let program = std::fs::read_to_string("/etc/os-release")
@@ -454,7 +468,7 @@ pub fn distro() -> Option<String> {
 
         match j.next()? {
             "PRETTY_NAME" => {
-                return Some(j.next()?.trim_matches('"').to_string())
+                return Some(j.next()?.trim_matches('"').to_string());
             }
             "NAME" => fallback = Some(j.next()?.trim_matches('"').to_string()),
             _ => {}
@@ -466,13 +480,13 @@ pub fn distro() -> Option<String> {
 
 #[cfg(target_os = "macos")]
 #[inline(always)]
-pub const fn desktop_env() -> DesktopEnv {
+pub(crate) const fn desktop_env() -> DesktopEnv {
     DesktopEnv::Aqua
 }
 
 #[cfg(not(target_os = "macos"))]
 #[inline(always)]
-pub fn desktop_env() -> DesktopEnv {
+pub(crate) fn desktop_env() -> DesktopEnv {
     match std::env::var_os("DESKTOP_SESSION")
         .map(|env| env.to_string_lossy().to_string())
     {
@@ -502,7 +516,7 @@ pub fn desktop_env() -> DesktopEnv {
 
 #[cfg(target_os = "macos")]
 #[inline(always)]
-pub const fn platform() -> Platform {
+pub(crate) const fn platform() -> Platform {
     Platform::MacOS
 }
 
@@ -512,10 +526,11 @@ pub const fn platform() -> Platform {
     target_os = "dragonfly",
     target_os = "bitrig",
     target_os = "openbsd",
-    target_os = "netbsd"
+    target_os = "netbsd",
+    target_os = "illumos"
 )))]
 #[inline(always)]
-pub const fn platform() -> Platform {
+pub(crate) const fn platform() -> Platform {
     Platform::Linux
 }
 
@@ -527,10 +542,16 @@ pub const fn platform() -> Platform {
     target_os = "netbsd"
 ))]
 #[inline(always)]
-pub const fn platform() -> Platform {
+pub(crate) const fn platform() -> Platform {
     Platform::Bsd
 }
 
+#[cfg(target_os = "illumos")]
+#[inline(always)]
+pub(crate) const fn platform() -> Platform {
+    Platform::Illumos
+}
+
 struct LangIter {
     array: String,
     index: Option<bool>,
@@ -540,31 +561,156 @@ impl Iterator for LangIter {
     type Item = String;
 
     fn next(&mut self) -> Option<Self::Item> {
-        if self.index? {
+        if self.index? && self.array.contains('-') {
             self.index = Some(false);
             let mut temp = self.array.split('-').next()?.to_string();
-            std::mem::swap(&mut temp, &mut self.array);
+            mem::swap(&mut temp, &mut self.array);
             Some(temp)
         } else {
             self.index = None;
             let mut temp = String::new();
-            std::mem::swap(&mut temp, &mut self.array);
+            mem::swap(&mut temp, &mut self.array);
             Some(temp)
         }
     }
 }
 
 #[inline(always)]
-pub fn lang() -> impl Iterator<Item = String> {
+pub(crate) fn lang() -> impl Iterator<Item = String> {
+    const DEFAULT_LANG: &str = "en_US";
+
     let array = std::env::var("LANG")
         .unwrap_or_default()
         .split('.')
         .next()
-        .unwrap_or("en_US")
-        .to_string()
-        .replace("_", "-");
+        .unwrap_or(DEFAULT_LANG)
+        .to_string();
+    let array = if array == "C" {
+        DEFAULT_LANG.to_string()
+    } else {
+        array
+    };
+
     LangIter {
-        array,
+        array: array.replace('_', "-"),
         index: Some(true),
     }
 }
+
+#[repr(C)]
+#[cfg(any(
+    target_os = "macos",
+    target_os = "ios",
+    target_os = "freebsd",
+    target_os = "dragonfly",
+    target_os = "netbsd",
+    target_os = "openbsd",
+    target_os = "illumos"
+))]
+struct UtsName {
+    #[cfg(not(target_os = "dragonfly"))]
+    sysname: [c_char; 256],
+    #[cfg(target_os = "dragonfly")]
+    sysname: [c_char; 32],
+    #[cfg(not(target_os = "dragonfly"))]
+    nodename: [c_char; 256],
+    #[cfg(target_os = "dragonfly")]
+    nodename: [c_char; 32],
+    #[cfg(not(target_os = "dragonfly"))]
+    release: [c_char; 256],
+    #[cfg(target_os = "dragonfly")]
+    release: [c_char; 32],
+    #[cfg(not(target_os = "dragonfly"))]
+    version: [c_char; 256],
+    #[cfg(target_os = "dragonfly")]
+    version: [c_char; 32],
+    #[cfg(not(target_os = "dragonfly"))]
+    machine: [c_char; 256],
+    #[cfg(target_os = "dragonfly")]
+    machine: [c_char; 32],
+}
+
+#[repr(C)]
+#[cfg(any(
+    target_os = "linux",
+    target_os = "android",
+    target_os = "fuchsia",
+    target_os = "redox"
+))]
+struct UtsName {
+    sysname: [c_char; 65],
+    nodename: [c_char; 65],
+    release: [c_char; 65],
+    version: [c_char; 65],
+    machine: [c_char; 65],
+    domainname: [c_char; 65],
+}
+
+// Buffer initialization
+impl Default for UtsName {
+    fn default() -> Self {
+        unsafe { mem::zeroed() }
+    }
+}
+
+extern "C" {
+    #[cfg(not(target_os = "freebsd"))]
+    fn uname(buf: *mut UtsName) -> c_int;
+
+    #[cfg(target_os = "freebsd")]
+    fn __xuname(nmln: c_int, buf: *mut c_void) -> c_int;
+}
+#[inline]
+#[cfg(target_os = "freebsd")]
+unsafe extern "C" fn uname(buf: *mut UtsName) -> c_int {
+    __xuname(256, buf.cast())
+}
+
+impl Arch {
+    fn from_str(s: Cow<'_, str>) -> Self {
+        let arch_str = match s {
+            Cow::Borrowed(str) => str,
+            Cow::Owned(ref str) => str,
+        };
+        match arch_str {
+            "aarch64" | "arm64" | "aarch64_be" | "armv8b" | "armv8l" => {
+                Arch::Arm64
+            }
+            "armv5" => Arch::ArmV5,
+            "armv6" | "arm" => Arch::ArmV6,
+            "armv7" => Arch::ArmV7,
+            "i386" => Arch::I386,
+            "i586" => Arch::I586,
+            "i686" | "i686-AT386" => Arch::I686,
+            "mips" => Arch::Mips,
+            "mipsel" => Arch::MipsEl,
+            "mips64" => Arch::Mips64,
+            "mips64el" => Arch::Mips64El,
+            "powerpc" | "ppc" | "ppcle" => Arch::PowerPc,
+            "powerpc64" | "ppc64" | "ppc64le" => Arch::PowerPc64,
+            "powerpc64le" => Arch::PowerPc64Le,
+            "riscv32" => Arch::Riscv32,
+            "riscv64" => Arch::Riscv64,
+            "s390x" => Arch::S390x,
+            "sparc" => Arch::Sparc,
+            "sparc64" => Arch::Sparc64,
+            "wasm32" => Arch::Wasm32,
+            "wasm64" => Arch::Wasm64,
+            "x86_64" | "amd64" => Arch::X64,
+            _ => Arch::Unknown(arch_str.to_owned()),
+        }
+    }
+}
+
+pub(crate) fn arch() -> Arch {
+    let mut buf = UtsName::default();
+    let result = unsafe { uname(&mut buf) };
+    if result == -1 {
+        return Arch::Unknown("uname(2) failed to execute".to_owned());
+    }
+
+    let arch_str =
+        unsafe { CStr::from_ptr(buf.machine.as_ptr()) }.to_string_lossy();
+
+    Arch::from_str(arch_str)
+}
